#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "vhdllex.h"
#include "konwert.h"
#include "stack.h"

/*
 * What does work faster:
 * 1. convertion string to lowercase before comparing strings
 * 2. comparing strings with case sensitivity turned off
 * ?
 */
#if 0
#define TO_LOWER_CONVERSION
#endif

#ifdef TO_LOWER_CONVERSION
#define STRCMP strcmp
#else
#define STRCMP stricmp
#endif

static stack_id idstack=NULL;

/*
 * Table of VHDL reserved words
 */
static char *reserved_words[]={"abs", "access", "after", "alias","all",
  "and","architecture","array","assert","attribute","begin","block",
  "body","buffer","bus","case","component","configuration","constant",
  "disconnect","downto","else","elsif","end","entity","exit","file","for",
  "function","generate","generic","group","guarded","if","impure","in",
  "inertial","inout","is","label","library","linkage","literal","loop",
  "map","mod","nand","new","next","nor","not","null","of","on","open",
  "or","others","out","package","port","postponed","procedure","process",
  "pure","range","record","register","reject","rem","report","return","rol",
  "ror","select","severity","signal","shared","sla","sll","sra","srl",
  "subtype","then","to","transport","type","unaffected","units","until",
  "use","variable","wait","when","while","with","xnor","xor"};

/*
 * Table of VHDL special characters
 */ 
static char special_characters[]={'\"','#','&','\'','(',')','*','+',
  ','/* comma */,'-','.'/* dot */,'/',':',';','<','=','>','_','|','!',
  '$','%','?','@','[','\\',']','^','`','{','}','~','\x0'/*inv exc. mark*/,
  '\x0'/*cent mark*/,'\x0'/*pound sign*/,'\x0'/*currency sign*/, 
  '\x0'/*yen sign*/,'\x0'/*broken bar*/,'\x0' /*paragraph sign*/,
  '\x0'/*diaeresis*/, '\x0'/*copyright sign*/,'\x0'/*feminin ordinal ind.*/,
  '\x0'/*left angle quot. mark*/, '\x0'/*not sign*/,'\x0'/*soft hyphen*/,
  '\x0'/*registered trade mark sign*/, '\x0'/*macron*/,'\x0'/*degree sign*/,
  '\x0'/*plus-minus sign*/, '\x0'/*superscript two*/,
  '\x0'/*superscript three*/,'\x0'/*acute accent*/,'\x0'/*micro sign*/,
  '\x0'/*pilcrow sign*/,'\x0'/*middle dot*/,'\x0'/*cedilla*/,
  '\x0'/*superscript one*/,'\x0'/*masculine ordinal indicator*/,
  '\x0'/*right angle quot. mark*/,'\x0'/*vulgar fraction one quarter*/,
  '\x0'/*vulgar fraction one half*/,'\x0'/*vulgar fraction three quarters*/,
  '\x0'/*inverted question mark*/,'\x0'/*multiplication mark*/,
  '\x0'/*division sign*/};

/*
 * Table of special double character symbols
 */
static char *double_char_symbols[]={"=>","**",":=","/=",">=","<=","<>"};

/*
 * Very unpleasent...
 * initialization is done to predefine types that should be
 * defined in std_logic_1164
 */
static int max_id_number=0;

/*
 * For local use only:
 */
int IsInList( char *, int *, int);
int AppendToList( char *);
char *bool2bit(char *);
char *ExcludeBitString(char *);
char *ExcludeString(char *);
char *ExcludeBasedInteger(char *);


int VHDL_ReservedWord( char *keyword)
  {
  int i;
  //printf("Inside VHDL_ReservedWord\n");
#ifdef TO_LOWER_CONVERSION
  for( i=0; keyword[i]!='\0'; i++)
    keyword[i]=tolower(keyword[i]);
#endif
  for( i=0; i<MAX_RESERVED_WORD_NUMBER; i++)
    if( !STRCMP( keyword, reserved_words[i]))
      return i+1;
  return -1;
  }

int VHDL_SpecialCharacters( char *key)
  {
  int i;
  //printf("Inside VHDL_SpecialCharacters\n");
  for( i=0; i<MAX_SPECIAL_CHARACTERS_NUMBER; i++)
    if( (*key)==special_characters[i])
      return i+FIRST_SPEC_CHAR_NUMBER;
  return -1;
  }

int VHDL_DoubleCharSymbols( char *symbol)
  {
  int i;
  //printf("Inside VHDL_DoubleCharSymbols\n");
  for( i=0; i<MAX_DOUBLE_CHAR_SYMBOL_NUMBER; i++)
    if( !strcmp(symbol,double_char_symbols[i]))
      return i+FIRST_DOUBLE_CHAR_SYMBOL_NUMBER;
  return -1;
  }

int VHDL_Identifier( char *id)
  {
  int number;

  //printf("Inside VHDL_Identifier\n");
  if( max_id_number)
    if( IsInList( id, &number, IGNORE_CASE))
      return number+FIRST_IDENTIFIER_NUMBER;
  number=AppendToList( id);
  return number+FIRST_IDENTIFIER_NUMBER;
  }

int VHDL_ExtendedIdentifier( char *id)
  {
  int number;

  //printf("Inside VHDL_ExtendedIdentifier\n");
  if( max_id_number)
    if( IsInList( id, &number, HOLD_CASE))
      return number+FIRST_IDENTIFIER_NUMBER;
  number=AppendToList( id);
  return number+FIRST_IDENTIFIER_NUMBER;
  }

int VHDL_Constant( char *id, int typeofconstant)
  {
  int lexnumber;
  int typenumber;
  char *bit_string, *tmp_id=NULL;

  //printf("Inside VHDL_Constant: %s\n", id);
  if( max_id_number)
    if( IsInList( id, &lexnumber, IGNORE_CASE))
      return lexnumber+FIRST_IDENTIFIER_NUMBER;

  //printf("Inside VHDL_Constant: po IsInList\n");
  lexnumber=AppendToList(id)+FIRST_IDENTIFIER_NUMBER;
  
  //printf("Inside VHDL_Constant: po AppendToList\n");
  if( typeofconstant==BIT_STRING_BIN ||
      typeofconstant==BIT_STRING_OCT ||
      typeofconstant==BIT_STRING_HEX )
    tmp_id=ExcludeBitString(id);
  if( typeofconstant==STRING_CONSTANT)
    tmp_id=ExcludeString(id);

  //printf("Inside VHDL_Constant: po ExcludeString\n");
  switch( typeofconstant)
    {
    case INTEGER_CONSTANT:       //printf("Przed anytobin\n"); fflush(stdout);
                                 bit_string=anytobin( id, 10); 
				 typenumber=201;
				 break;
    case BIT_STRING_BIN:         bit_string=bitbintobin( tmp_id);
				 typenumber=203;
				 break;
    case BIT_STRING_OCT:         bit_string=bitocttobin( tmp_id);
				 typenumber=203;
				 break;
    case BIT_STRING_HEX:         bit_string=bithextobin( tmp_id);
				 typenumber=203;
				 break;
    case CHARACTER_CONSTANT:   
                                 //printf("Alokacja\n");
#if 0
                                 bit_string=(char*)malloc(strlen(id));
				 strcpy(bit_string,id);
#else
				 bit_string=strdup(id);
#endif
				 typenumber=205;
				 break;
    case INTEGER_BASED_CONSTANT: bit_string=ExcludeBasedInteger(id);
				 typenumber=201;
				 break;
    case STRING_CONSTANT:        
#if 0
				 bit_string=(char *)malloc(strlen(tmp_id));
    			         //printf("Alokacja\n");
				 strcpy(bit_string,tmp_id);
#else
				 bit_string=strdup(tmp_id);
#endif
				 typenumber=206;
				 break;
    case PREDEFINED_ATTRIBUTE:   typenumber=0;
    			         //printf("Alokacja\n");
				 bit_string=(char *)malloc(1);
				 bit_string[0]='\0';
				 break;
    case BOOLEAN_CONSTANT:       typenumber=204;
				 bit_string=bool2bit(id);
				 break;
    case EDGE_FUNCTIONS:         typenumber=204;
				 bit_string=(char *)malloc(1);
				 bit_string[0]='\0';
				 break;
    default: return lexnumber;
    }
  //printf("Inside VHDL_Constant: po switch\n");
  fprintf( consfile, "%d\t%s\t%d\t5\t0\t%d\t0\t%d\t%s\n",
	   lexnumber, id, typenumber, (int)strlen(bit_string), 
	   (int)strlen(bit_string)-1, bit_string);
  fflush(consfile);
  if(bit_string)
    free(bit_string);
  if(tmp_id!=NULL)
    free(tmp_id);
  return lexnumber;
  }

int IsInList( char *id, int *number, int ignore)
  {
  if(stack_find(&idstack,number,id,ignore)==OK)
    return 1;
  return 0;
  }

int AppendToList( char *id)
  {
  fprintf( idfile, "%d\t%s\n",++max_id_number,id);
  fflush(idfile);
  stack_push(&idstack,max_id_number-1,id);
  return max_id_number-1;
  }

char *ExcludeBitString(char *id)
  {
  char *result=(char *)malloc(strlen(id));
  char *tmp=strchr(id,'"');
  //printf("Alokacja\n");
  strcpy(result,++tmp);
  tmp=strchr(result,'"');
  (*tmp)='\0';
  return result;
  }

char *ExcludeString(char *id)
  {
  char *result=(char *)malloc(strlen(id));
  char *tmp=strchr(id,'"');
  //printf("Alokacja\n");
  strcpy(result,++tmp);
  result[strlen(id)-2]='\0';
  return result;
  }

char *ExcludeBasedInteger(char *id)
  {
  char *result, tmp[256];
  char based_number[256];
  char expstr[256];
  int base=0, exponent=0;
  int i, j;
  int position;
  for( i=0; id[i]!='\0'; i++)
    if( id[i] != '#')
      tmp[i]=id[i];
    else
      {
      tmp[i++]='\0';
      break;
      }
  sscanf(tmp,"%d",&base);

  for( j=0; id[j+i]!='\0'; j++)
    if( id[j+i]!='#')
      tmp[j]=id[j+i];
    else
      {
      tmp[j++]='\0';
      position=j-1; /* position points to last character in tmp string */
      break;
      }
  strcpy( based_number, tmp);

/*
 * Is there an exponent?
 */
  if( id[j+i] != '\0')
    {
    char *tmpstr;
    i=j+i;
    for( j=0; id[i+j]!='\0'; j++)
      expstr[j]=id[i+j];
    expstr[j]='\0';
    /* expstr contains exponent, including 'e' and (maybe) '+' */
    for( tmpstr=expstr; !isdigit(*tmpstr); tmpstr++)
      ;
    sscanf( tmpstr,"%d",&exponent);
    }

  for( j=0; j<exponent; j++)
    based_number[position+j]='0';
  based_number[position+j]='\0';
  return result=anytobin( based_number, base);
  }

void InsertPredefinedAttributes(void)
  {
  char *predefined_attributes[]={ "'base", "'left", "'right", "'high", "'low",
				  "'range", "'reverse_range", "'length",
				  "'stable", "'event", "'unstable", "'ascending", 
                                  "'descending", NULL };
  int i;
  for( i=0; predefined_attributes[i]; i++)
    VHDL_Constant( predefined_attributes[i], PREDEFINED_ATTRIBUTE);
  }

void InsertPredefinedConstants(void)
  {
  char *predefined_constants[]={ "false", "true", NULL};
  int i;
  for( i=0; predefined_constants[i]; i++)
    VHDL_Constant( predefined_constants[i], BOOLEAN_CONSTANT);
  }

void InsertPredefinedEdgeFuctions(void)
  {
  char *predefined_constants[]={ "rising_edge", "falling_edge", NULL};
  int i;
  for( i=0; predefined_constants[i]; i++)
    VHDL_Constant( predefined_constants[i], EDGE_FUNCTIONS);
  }

void InsertPredefinedIds(void)
  {
  char *predefined_ids[]={ "integer","bit","bit_vector",
       "boolean","character","string","natural","positive",
       "std_ulogic","std_logic","std_ulogic_vector",
       "std_logic_vector","signed","unsigned","small_int",
       "UX01", "UX01Z", "X01Z", "X01", NULL};
  int i;
  for( i=0; predefined_ids[i]; i++)
    {
    //printf("%s\n", predefined_ids[i]);
    VHDL_Identifier( predefined_ids[i]);
    }
  }

char *bool2bit( char *id)
  {
  char *result=(char*)malloc(2);
  //printf("Alokacja\n");
  if( !stricmp(id,"false"))
    {
    strcpy(result,"0");
    return result;
    }
  if( !stricmp(id,"true"))
    {
    strcpy(result,"1");
    return result;
    }
  return NULL;
  }

void GetLastLexemNumber(void)
  {
  FILE *lastlexfile=fopen("..\\last.no","rt");
  
  if(!lastlexfile)
    if( !(lastlexfile=fopen("..\\syslib\\last.no","rt")))
      return /*void*/;
  if(!fscanf(lastlexfile,"%d",&max_id_number))
    {
    fprintf(stderr,"Corrupt syslib/last.no file.\n");
    exit(-1);
    }
  fclose(lastlexfile);
  return /*void*/;
  }

int PutLastLexemNumber(void)
  {
  FILE *lastlexfile=fopen("last.no","wt");
  if(!lastlexfile)
    return -1;
  fprintf(lastlexfile,"%d\n",max_id_number);
  fclose(lastlexfile);
  return 0;
  }

int InsertPredefinedCharacterConstants(void)
  {
  int i;
  char *control_characters[]={ "nul", "soh", "stx", "etx", "eot", "enq", "ack", "bel",
                               "bs",  "ht",  "lf",  "vt",  "ff",  "cr",  "so",  "si",
                               "dle", "dc1", "dc2", "dc3", "dc4", "nak", "syn", "etb",
                               "can", "em",  "sub", "esc", "fsp", "gsp", "rsp", "usp",
                               NULL};
  char *higher_characters[]={ "del",  "c128", "c129", "c130", "c131", "c132", "c133", "c134",
                              "c135", "c136", "c137", "c138", "c139", "c140", "c141", "c142",
                              "c143", "c144", "c145", "c146", "c147", "c148", "c149", "c150",
                              "c151", "c152", "c153", "c154", "c155", "c156", "c157", "c158",
                              "c159", NULL};
/*Dodal Marcin Liersz*/
return 0;
/*Tak zapisane znaki powoduja brak wykonywania LPMow, gdzie so, si .. 
sa identyfikatorami, a blad wychodzi dopiero przy generatorze kodu */

  for( i=0; control_characters[i]!=NULL; i++)
    {
    char *str=chartobin(i);
    int lexnumber=AppendToList(control_characters[i])+FIRST_IDENTIFIER_NUMBER;
    fprintf( consfile, "%d\t%s\t%d\t5\t0\t%d\t0\t%d\t%s\n",
             lexnumber, control_characters[i], 205, 8, 255, str);
    free(str);
    }
  for( i=0; higher_characters[i]!=NULL; i++)
    {
    char *str=chartobin(i+127);
    int lexnumber=AppendToList(higher_characters[i])+FIRST_IDENTIFIER_NUMBER;
    fprintf( consfile, "%d\t%s\t%d\t5\t0\t%d\t0\t%d\t%s\n",
             lexnumber, higher_characters[i], 205, 8, 255, str);
    free(str);
    }
  return 0;
  }
